'use strict'
var __importStar =
	(this && this.__importStar) ||
	function (mod) {
		if (mod && mod.__esModule) return mod
		var result = {}
		if (mod != null)
			for (var k in mod)
				if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k]
		result['default'] = mod
		return result
	}
var __importDefault =
	(this && this.__importDefault) ||
	function (mod) {
		return mod && mod.__esModule ? mod : { default: mod }
	}
Object.defineProperty(exports, '__esModule', { value: true })
const extendr_1 = require('extendr')
const stream_1 = require('stream')
const rfc_log_levels_1 = __importStar(require('rfc-log-levels'))
const get_current_line_1 = __importDefault(require('get-current-line'))
/**
 * Logger.
 * This is what we write to.
 * It extends from PassThrough and not transform.
 * If you are piping / writing directly to the logger, make sure it corresponds to the correct entry format (as described in `log`).
 *
 * @param args forwarded to {@link Logger#setConfig}
 *
 * @example Creation
 * ``` javascript
 * // Via class
 * import { Logger } from 'caterpillar'
 * const logger = new Logger()
 * // Via create helper
 * const logger = Logger.create()
 * // Via create alias
 * const logger = require('caterpillar').create()
 * ```
 */
class Logger extends stream_1.PassThrough {
	constructor(...args) {
		super(...args)
		this._config = this.getInitialConfig()
		this.setConfig(...args)
	}
	// ===================================
	// Generic Differences
	// This code is shared but different between Logger and Transform
	/**
	 * Get the initial configuration.
	 * Initial log levels are fetched from: https://github.com/bevry/rfc-log-levels
	 */
	getInitialConfig() {
		return {
			lineOffset: {
				file: __filename,
				method: /log/i,
			},
			levels: Object.assign({}, rfc_log_levels_1.rfcLogLevels, {
				default: 6,
			}),
		}
	}
	/**
	 * Alternative way of creating an instance of the class without having to use the `new` keyword.
	 * Useful when creating the class directly from `require` statements.
	 */
	static create(...args) {
		return new this(...args)
	}
	// ===================================
	// Generic
	// This code is shared between Logger and Transform
	/**
	 * Get the current configuration object for this instance.
	 */
	getConfig() {
		return this._config
	}
	/**
	 * Apply the specified configurations to this instance's configuration via deep merging.
	 * @example
	 * ``` javascript
	 * setConfig({a: 1}, {b: 2})
	 * getConfig()  // {a: 1, b: 2}
	 * ```
	 */
	setConfig(...configs) {
		extendr_1.deep(this._config, ...configs)
		this.emit('config', ...configs)
		return this
	}
	/**
	 * Pipe this data to some other writable stream.
	 * If the child stream also has a `setConfig` method, we will ensure the childs configuration is kept consistent with parents.
	 * @param child stream to be piped to
	 * @returns the result of the pipe operation
	 */
	pipe(child) {
		if (typeof child.setConfig !== 'undefined') {
			child.setConfig(this.getConfig())
			const listener = child.setConfig.bind(child)
			this.on('config', listener)
			child.once('close', () => this.removeListener('config', listener))
		}
		return super.pipe(child)
	}
	// ===================================
	// Logger
	/**
	 * Takes an arguments array and tranforms it into a log entry
	 * @param args
	 */
	getLogEntry(args) {
		const { lineOffset, levels } = this.getConfig()
		const date = new Date().toISOString()
		const lineInfo = get_current_line_1.default(lineOffset)
		const level = args.shift()
		let levelInfo = rfc_log_levels_1.default(level, levels)
		if (levelInfo == null) {
			levelInfo = rfc_log_levels_1.default('default', levels)
			if (levelInfo == null) {
				throw new Error(
					'caterpillar: rfc-log-levels: no default log level configuration was provided'
				)
			}
			args.unshift(level)
		}
		return Object.assign({ date, args }, levelInfo, lineInfo)
	}
	/**
	 * Log the arguments into the logger stream as formatted data with debugging information.
	 * Such that our transformers can deal with it intelligently.
	 *
	 * @param args forwarded to {@link Logger#getLogEntry}
	 *
	 * @example Inputs
	 * ``` javascript
	 * logger.log('note', 'this is working swell')
	 * ```
	 * ``` javascript
	 * logger.log('this', 'worked', 'swell')
	 * ```
	 *
	 * @example Results
	 * ``` json
	 * {
	 * 	"args": ["this is working swell"],
	 * 	"date": "2013-04-25T10:18:25.722Z",
	 * 	"levelNumber": 5,
	 * 	"levelName": "notice",
	 * 	"line": "59",
	 * 	"method": "Object.<anonymous>",
	 * 	"file": "/Users/balupton/some-project/calling-file.js"
	 * }
	 * ```
	 * ``` json
	 * {
	 *		"args": ["this", "worked", "well"],
	 *		"date": "2013-04-25T10:18:26.539Z",
	 *		"levelNumber": 6,
	 *		"levelName": "info",
	 *		"line": "60",
	 *		"method": "Object.<anonymous>",
	 *		"file": "/Users/balupton/some-project/calling-file.js"
	 * }
	 * ```
	 */
	log(...args) {
		// Fetch the log entry
		const entry = this.getLogEntry(args)
		// Write the arguments as an entry to be transformed by our format
		this.write(JSON.stringify(entry))
		// Chain
		return this
	}
}
exports.default = Logger
